'''Handles all map operations'''

import bge
import logging
import random
import math

import Common
import defs
import mathutils
import Loader
import Economy
import DynTex
import UI
import Buildings


LOG = logging.getLogger()
SEED = 10


def generate_list_of_lists(size):
    '''Generates a list of lists of a specified size, duh'''
    tmp = list()
    for row in range(0, size):
        tmp.append(list())
        for column in range(0, size):
            tmp[row].append(list())
            tmp[row][column] = 0
    return tmp


def generate_map(cont):
    if 'map' not in generate_map.__dict__:
        bge.logic.map = Map(cont)
        generate_map.map = bge.logic.map

    function_list = bge.logic.map.load_functions
    total = len(function_list)
    count = generate_map.__dict__.get('count', 0)
    if count < total:
        Loader.display_load_message('Generating Map {}/{}'.format(count, total))
        function_list[count](cont)
        LOG.debug("Function: %s", function_list[count])
        generate_map.count = count + 1
        return False


class Map(object):
    '''Contains the map!'''
    def __init__(self, cont, seed=SEED):
        mathutils.noise.seed_set(seed)
        self.size = defs.WORLD_SIZE

        self.load_functions = [
            self.generate_rock,
            self.generate_dirt,
            self.generate_bedrock,
            self.add_floor,
            self.place_entry,
            self.add_crystals,
            self.initial_place_tiles,
        ]

        self.raw_map = generate_list_of_lists(self.size)
        self.obj_map = generate_list_of_lists(self.size)
        self.placables_map = generate_list_of_lists(self.size)

    def place_entry(self, cont):
        '''Places the mine entry'''
        iters = 0
        while iters < 2000:
            rand_x = int(mathutils.noise.random() * defs.WORLD_SIZE)
            rand_y = int(mathutils.noise.random() * defs.WORLD_SIZE)
            if self.get_shape(rand_x, rand_y) == [[0, 0], [0, 0]]:
                position = [rand_x, rand_y]
                Buildings.place_entry(node_to_world(position))
                LOG.debug("Placed base in %d iterations", iters)
                return
            iters += 1

        LOG.error("Unable to place base nicely. Apply brute force")
        center = int(defs.WORLD_SIZE / 2)
        self.raw_map[center][center] = 0
        self.raw_map[center + 1][center] = 0
        self.raw_map[center][center + 1] = 0
        self.raw_map[center + 1][center + 1] = 0

        Buildings.place_entry(node_to_world([center, center]))


    def add_crystals(self, cont):
        '''Places crystals around the map'''
        placed = 0
        iters = 0
        while placed < defs.NUM_CRYSTAL_ATTEMPTS and iters < 2000:
            rand_x = int(mathutils.noise.random() * defs.WORLD_SIZE)
            rand_y = int(mathutils.noise.random() * defs.WORLD_SIZE)
            if self.raw_map[rand_x][rand_y] == 0 and \
                    self.get_placable(rand_x, rand_y) == 0:
                x_pos = self.size / 2 - rand_x + 0.5
                y_pos = self.size / 2 - rand_y + 0.5
                crystal = add_crystal(cont, x_pos, y_pos)
                self.add_placable(rand_x, rand_y, crystal)
                placed += 1
            iters += 1

    def add_placable(self, x_pos, y_pos, obj):
        '''Adds a placable to the map'''
        self.placables_map[x_pos][y_pos] = obj

    def get_placable(self, x_pos, y_pos):
        '''Gets the placable at a point in the map or 0 if nothing is
        there'''
        return self.placables_map[x_pos][y_pos]

    def add_floor(self, cont):
        '''Adds the cavern floor'''
        scene = cont.owner.scene
        floor = scene.addObject('Floor', scene.objects[0])
        floor.worldPosition = [0, 0, 0]
        floor.worldScale = [defs.WORLD_SIZE + 1] * 2 + [1]
        floor.worldOrientation = [0, 0, 0]
        self.floor = floor
        self.world_tex = DynTex.DynTex(self.floor, defs.WORLD_SIZE)

    def generate_bedrock(self, cont):
        '''Lays down the toughest material: bedrock'''
        for x_raw in range(self.size):
            for y_raw in range(self.size):
                x_pos = x_raw * defs.BEDROCK_SCALE
                y_pos = y_raw * defs.BEDROCK_SCALE
                value = mathutils.noise.noise([x_pos, y_pos, SEED])
                if value > defs.BEDROCK_THRESHOLD:
                    self.raw_map[x_raw][y_raw] = 5

    def generate_rock(self, cont):
        '''Places the metal deposits'''
        for x_raw in range(self.size):
            for y_raw in range(self.size):
                value = mathutils.noise.random()
                if value > defs.METAL_THRESHOLD:
                    self.raw_map[x_raw][y_raw] = 2
                else:
                    self.raw_map[x_raw][y_raw] = 0
        self.amazing_algorithm()
        self.amazing_algorithm()

    def generate_dirt(self, cont):
        '''Fills in some dirt'''
        self.erode()

    def erode(self):
        '''Performans a binary erode on the map'''
        tmp_map = generate_list_of_lists(self.size)
        for x_raw in range(self.size):
            for y_raw in range(self.size):
                right = self.get_point(x_raw + 1, y_raw)
                left = self.get_point(x_raw - 1, y_raw)
                top = self.get_point(x_raw, y_raw + 1)
                bottom = self.get_point(x_raw, y_raw - 1)
                val = max(top, bottom, left, right)
                tmp_map[x_raw][y_raw] = int(val / 2)
        for x_raw in range(self.size):
            for y_raw in range(self.size):
                if tmp_map[x_raw][y_raw] > self.raw_map[x_raw][y_raw]:
                    self.raw_map[x_raw][y_raw] = tmp_map[x_raw][y_raw]

    def amazing_algorithm(self):
        '''Does the CaveX14 algorithm....'''
        for x_raw in range(self.size):
            for y_raw in range(self.size):
                right = self.get_point(x_raw + 1, y_raw)
                left = self.get_point(x_raw - 1, y_raw)
                top = self.get_point(x_raw, y_raw + 1)
                bottom = self.get_point(x_raw, y_raw - 1)
                if top == bottom and left == right:
                    self.raw_map[x_raw][y_raw] = 0

    def get_point(self, x_pos, y_pos):
        '''Returns the health of the rock at the specified location'''
        if x_pos >= self.size or x_pos < 0 or y_pos >= self.size or y_pos < 0:
            return 9
        else:
            return self.raw_map[x_pos][y_pos]

    def __str__(self):
        out_str = ''
        for x_raw in range(self.size):
            for y_raw in range(self.size):
                value = self.get_point(x_raw, y_raw)
                out_str += VALUES[int(value)] + ' '
            out_str += '\n'
        return out_str

    def initial_place_tiles(self, cont):
        '''Makes the map into something visible to the player'''

        for x_raw in range(self.size):
            for y_raw in range(self.size):
                pos = [self.size / 2 - x_raw, self.size / 2 - y_raw]
                shape = self.get_shape(x_raw, y_raw)
                added = add_tile(cont, shape, pos)
                if added is not None:
                    added['pos'] = (x_raw, y_raw)
                self.obj_map[x_raw][y_raw] = added
                self.update_dyn_tex(x_raw, y_raw)

    def update_dyn_tex(self, x_raw, y_raw):
        '''Updates the texture that shows the map status'''
        val = self.raw_map[x_raw][y_raw]
        health = int(val / 10 * 255)
        occluded = 255 if val != 0 else 0
        col = [health, occluded, 0]
        self.world_tex.set_pixel(x_raw, y_raw, col + [255])

    def set_point(self, x_raw, y_raw, value):
        '''Sets a point including changing blender models'''
        if x_raw < 0 or x_raw >= self.size:
            return
        if y_raw < 0 or y_raw >= self.size:
            return
        prev = math.ceil(self.get_point(x_raw, y_raw))
        self.raw_map[x_raw][y_raw] = value
        self.update_dyn_tex(x_raw, y_raw)

        if prev != 0 and value == 0:
            for delta_x in range(-1, 1):
                for delta_y in range(-1, 1):
                    x_pos = x_raw + delta_x
                    y_pos = y_raw + delta_y
                    old_obj = self.obj_map[x_pos][y_pos]
                    if old_obj is not None:
                        old_obj.endObject()

                    shape = self.get_shape(x_pos, y_pos)
                    pos = [self.size / 2 - x_pos, self.size / 2 - y_pos]
                    cont = bge.logic.getCurrentController()
                    added = add_tile(cont, shape, pos)
                    if added is not None:
                        added['pos'] = (x_pos, y_pos)
                    self.obj_map[x_pos][y_pos] = added

    def mine(self, x_raw, y_raw, multiplier, add_directly=True):
        '''Mines a node slowly'''
        mine_quantity = defs.MINE_RATE / bge.logic.getAverageFrameRate()
        mine_quantity *= multiplier

        prev_amount = self.get_point(x_raw, y_raw)
        amount = prev_amount - mine_quantity
        self.set_point(x_raw, y_raw, max(0, amount))

        total_value = 0
        for val_range in defs.ORE_GATHER_RATE:
            start_val, stop_val = val_range
            if amount > start_val and amount < stop_val:
                money_value = defs.ORE_GATHER_RATE[val_range] * mine_quantity
                total_value += money_value
                if add_directly is True:
                    Economy.get_wallet().add_resource({'Ore': money_value})
        return total_value

    def get_shape(self, x_pos, y_pos):
        '''Returns the shape of four points near a node'''
        shape = [[self.get_point(x_pos, y_pos + 1),
                  self.get_point(x_pos + 1, y_pos + 1)],
                 [self.get_point(x_pos, y_pos),
                  self.get_point(x_pos + 1, y_pos)]]
        return shape


def get_tile_type(name):
    '''Returns a random tile of the correct type'''
    scene = Common.get_scene(defs.GAME_SCENE_NAME)
    all_of_type = [o for o in scene.objectsInactive if o.name.startswith(name)]
    return random.choice(all_of_type)


def add_crystal(cont, x_pos, y_pos):
    '''Adds a crystal'''
    scene = cont.owner.scene
    crystal = scene.addObject('Crystal', cont.owner)
    crystal.worldPosition.x = x_pos
    crystal.worldPosition.y = y_pos
    crystal.worldPosition.z = 0
    crystal.worldScale = [1, 1, 1]
    crystal.worldOrientation = [0, 0, 0]
    return crystal


def add_tile(cont, shape, position):
    '''Adds a specific shape tile in teh specified position. Examples
    of shapes are [[1,0][0,0]], [[0,1],[0,1]]'''
    scene = cont.owner.scene

    shape_dict = COMPLETE_SHAPE_DICT
    shape = list(list(x) for x in shape)
    for row in [0, 1]:
        for col in [0, 1]:
            if shape[row][col] != 0:
                shape[row][col] = 1
    hashable_shape = tuple(tuple(x) for x in shape)
    if hashable_shape in shape_dict:
        shape_name, num_rots = shape_dict[hashable_shape]
        if shape_name not in ['Floor', 'Roof']:
            tile_type = get_tile_type(shape_name)
            new_tile = scene.addObject(tile_type, scene.objects[0])

            new_tile.worldScale = [1, 1, 1]
            new_tile.worldPosition.xy = position
            new_tile.worldPosition.z = 0
            new_tile.worldOrientation = [0, 0, num_rots * math.pi / 2]

            return new_tile
    else:
        LOG.error("Attempted to add %s", str(hashable_shape))


def build_complete_shape_dict():
    '''Reverses SHAPE_DICT so that the object can be found by
    looking up the desired shape'''
    def rotate_shape(shape, num):
        '''Rotates a shape so that you only need one orientation of
        an object in mesh form'''
        new_shape = list(list(x) for x in shape)
        for _ in range(num + 1):
            tmp = new_shape[0][0]
            new_shape[0][0] = new_shape[0][1]
            new_shape[0][1] = new_shape[1][1]
            new_shape[1][1] = new_shape[1][0]
            new_shape[1][0] = tmp
        return tuple(tuple(x) for x in new_shape)

    new_shape_dict = {}
    for raw_shape in SHAPE_DICT:
        for i in range(4):
            shape = rotate_shape(SHAPE_DICT[raw_shape], i)

            new_shape_dict[shape] = [raw_shape, i]

    return new_shape_dict


def get_closest_node(pos, inter=1):
    '''Returns the coordinates of the closest node'''
    # Parts are placed using [size / 2 - x_pos, size / 2 - y_pos]
    # So we have to do the inverse
    x_pos = defs.WORLD_SIZE / 2 - pos[0]
    y_pos = defs.WORLD_SIZE / 2 - pos[1]
    if inter == 1:
        return [math.ceil(x_pos), math.ceil(y_pos)]
    else:
        return [x_pos, y_pos]


def node_to_world(pos):
    '''Converts from node co-ordinates to blender world coordinates'''
    x_pos = defs.WORLD_SIZE / 2 - pos[0] + 0.5
    y_pos = defs.WORLD_SIZE / 2 - pos[1] + 0.5
    return [x_pos, y_pos, 0]


VALUES = ['  ', '..', '||', '[]', '$$', '##']
SHAPE_DICT = {
    'Floor': ((0, 0),
              (0, 0)),

    'Wall': ((0, 1),
             (0, 1)),

    'Split': ((0, 1),
              (1, 0)),

    'Inner Corner': ((0, 1),
                     (1, 1)),

    'Outer Corner': ((0, 0),
                     (0, 1)),

    'Roof': ((1, 1),
             (1, 1)),
}

COMPLETE_SHAPE_DICT = build_complete_shape_dict()
