'''Handles unit operations'''
import bge
import mathutils

import logging
import random

import Map
import defs
import Mouse
import Control
import Buildings

LOG = logging.getLogger()


def init_units(_):
    '''Adds the units to the scheduler'''
    bge.logic.scheduler.register_function(update_units)
    bge.logic.unit_list = list()


def update_units():
    '''Runs every frame to update the path of each unit'''
    for unit in bge.logic.unit_list:
        unit.update()


def assign_target(event, pos, selected, hover):
    '''Assigns the target for all selected units'''
    if event is None:
        return None
    mouse_node = pos
    if event[defs.MOUSE_MAP['MOVE']] != 0:
        for unit in bge.logic.unit_list:
            if unit.selected:
                unit.set_target_node(mouse_node)

    if event[defs.MOUSE_MAP['SELECT']] == 1:
        if Control.get_key('LEFTSHIFTKEY') == 0:
            for unit in bge.logic.unit_list:
                unit.select(False)
            Mouse.set_selected(None)
        select_unit(event, pos, selected, hover)


def deselect_all_units():
    '''Deselects all units'''
    for unit in bge.logic.unit_list:
        unit.select(False)


def select_unit(event, pos, selected, hover):
    '''Selects the unit under the mouse'''
    if event is None:
        return

    if event[defs.MOUSE_MAP['SELECT']] == bge.logic.KX_SENSOR_JUST_ACTIVATED:
        if isinstance(hover, Unit):
            hover.select(True)
            Mouse.set_selected(assign_target)


def add_unit(obj_type, spawn_obj):
    '''Spawns the unit at the supplied object and mutates it'''
    obj_name = obj_type['obj_name']
    level = obj_type['level']
    obj = spawn_obj.scene.objectsInactive[obj_name]
    new_unit = spawn_obj.scene.addObject(obj, spawn_obj)
    new_unit = UNIT_MUTATE_DICT[obj_type['unit_type']](new_unit, level)
    bge.logic.unit_list.append(new_unit)


class Unit(bge.types.KX_GameObject):
    '''Base class for a generic unit'''
    def __init__(self, _):
        self.selected = False
        self.speed = 0
        self.target_node = None
        self.direction_indicator = DirectionIndicator()
        self.definite_target_node = None

    def set_text(self, designation):
        '''Specifies a unit's identifier'''
        for child in self.childrenRecursive:
            if isinstance(child, bge.types.KX_FontObject):
                child.resolution = 3
                child.text = designation

    def select(self, state):
        '''Sets if a unit is selected or not'''
        self.selected = state

    def set_target_node(self, node):
        '''Sets the position the vehicle is travelling towards'''
        self.target_node = node
        self.definite_target_node = node

    def move(self, vec):
        '''Makes the unit attempt to move along target vector'''
        if vec.length > 0.1:
            vec.length = min(vec.length * 4, self.speed)
        else:
            self.target_node = None
            return
        vec3 = list(vec) + [0]
        # self.worldLinearVelocity.xy = vec
        self.localLinearVelocity.y = vec.length
        self.localLinearVelocity.x = 0
        self.face_direction(vec3)
        if self.worldLinearVelocity.z > 0.1:
            self.worldLinearVelocity.z = 0.001

    def get_terrain(self, cardinal_delta=None):
        '''Returns the terrain the rover is on or in the direction
        it is trying to move if supplied'''
        if cardinal_delta is None:
            cardinal_delta = [0, 0]
        x_pos, y_pos = Map.get_closest_node(self.worldPosition)
        terrain = bge.logic.map.get_point(x_pos, y_pos)
        if terrain == 0:
            x_pos = x_pos - cardinal_delta[0]
            y_pos = y_pos - cardinal_delta[1]
            terrain = bge.logic.map.get_point(x_pos, y_pos)
        return [terrain, x_pos, y_pos]

    def update_direction_indicator(self):
        '''Updates the units direction indicator to show it's
        current heading'''
        if self.target_node is not None:
            from_pos = self.worldPosition.copy()
            from_pos.z = 0.1
            to_pos = mathutils.Vector(Map.node_to_world(self.target_node))
            to_pos += mathutils.Vector([-0.5, -0.5, 0])
            to_pos.z = 0.1
            self.direction_indicator.set_direction(from_pos, to_pos)

    def face_direction(self, direction):
        '''Makes the unit face in the specified direction'''
        direction = mathutils.Vector(direction)
        if direction[0] != 0 or direction[1] != 0:
            self.alignAxisToVect(direction, 1, 0.2)
            self.alignAxisToVect([0, 0, 1], 2)

            delta = direction.normalized() - self.getAxisVect([0, -1, 0])
            if delta.length < 0.01:  # Want to go the opposite direction
                self.applyRotation([0, 0, 0.1], False)


class MinerUnit(Unit):
    '''The level 1 mining unit'''

    def __init__(self, obj, level):
        super().__init__(obj)
        self.set_text('M{:02d}'.format(random.randint(0, 99)))
        self.collected_resources = {'Ore': 0, 'Crystal': 0}

        self.specs = defs.MINING_VEHICLE_LEVELS[level]

    def check_placables(self):
        '''Interacts with placables at current position'''
        x_pos, y_pos = Map.get_closest_node(self.worldPosition)
        placable = bge.logic.map.get_placable(x_pos, y_pos)
        if isinstance(placable, bge.types.KX_GameObject):
            if placable.get('Droppable', False):
                placable.add_resources(self.collected_resources)
                for resource in self.collected_resources:
                    self.collected_resources[resource] = 0
                    self.target_node = self.definite_target_node
            if placable.get('SmoothSurface', False) is not False:
                self.speed = self.specs['smooth_speed']
            else:
                self.speed = self.specs['rough_speed']

        else:
            self.speed = self.specs['rough_speed']

    def mine_map(self, pos, direction):
        '''Mines the specified position and faces the specified direction'''
        self.face_direction(direction + [0])
        if self.get_resources_quantity() < self.specs['resource_limit']['Ore']:
            collected = bge.logic.map.mine(pos[0], pos[1], 1, False)
            self.collected_resources['Ore'] += collected
            return True
        else:
            self.target_node = mathutils.Vector(get_nearest_dropsite(pos))
            return False

    def get_resources_quantity(self):
        '''Returns the total count of resources stored in the unit'''
        total = 0
        for resource in self.collected_resources:
            total += self.collected_resources[resource]
        return total

    def update(self):
        '''Update this units mining, pathfinding etc.'''
        if self.target_node is not None:
            unit_pos = Map.get_closest_node(self.worldPosition, inter=0)
            if unit_pos != self.target_node:

                delta = mathutils.Vector(unit_pos) - \
                    mathutils.Vector(self.target_node)
                cardinal_delta = convert_to_cardinal(delta)

                terrain, x_pos, y_pos = self.get_terrain(cardinal_delta)

                if terrain == 0:  # Open space
                    if cardinal_delta[0] == 0 and cardinal_delta[1] == 0:
                        self.move(delta)  # Closer than a node distance
                    else:
                        self.move(mathutils.Vector(cardinal_delta))

                    self.check_placables()

                else:  # Something there
                    pos = [x_pos, y_pos]
                    if not self.mine_map(pos, cardinal_delta):
                        print("I'm Stuck because I don't know pathfinding")

        if self.selected:
            if self.target_node is not None:
                self.direction_indicator.set_visible(True)
                self.update_direction_indicator()
            else:
                self.direction_indicator.set_visible(False)
        else:
            self.direction_indicator.set_visible(False)


class ResourceUnit(Unit):
    '''A unit that allows mining vehicles to deposit resources into
    itself so they don't have to travel as far'''
    def __init__(self, obj, level):
        super().__init__(obj)
        self.set_text('M{:02d}'.format(random.randint(0, 99)))
        self.collected_resources = {'Ore': 0, 'Crystal': 0}

        self.specs = defs.RESOURCE_VEHILCE_LEVELS[level]

    def check_placables(self):
        '''Interacts with placables at current position'''
        x_pos, y_pos = Map.get_closest_node(self.worldPosition)
        placable = bge.logic.map.get_placable(x_pos, y_pos)
        if isinstance(placable, bge.types.KX_GameObject):
            if placable.get('Droppable', False):
                placable.add_resources(self.collected_resources)
                for resource in self.collected_resources:
                    self.collected_resources[resource] = 0
                    self.target_node = self.definite_target_node
            if placable.get('SmoothSurface', False) is not False:
                self.speed = self.specs['smooth_speed']
            else:
                self.speed = self.specs['rough_speed']

        else:
            self.speed = self.specs['rough_speed']

    def update(self):
        '''Runs every time the resource truck needs to be updated'''
        if self.target_node is not None:
            unit_pos = Map.get_closest_node(self.worldPosition, inter=0)
            if unit_pos != self.target_node:

                delta = mathutils.Vector(unit_pos) - \
                    mathutils.Vector(self.target_node)
                cardinal_delta = convert_to_cardinal(delta)

                terrain, x_pos, y_pos = self.get_terrain(cardinal_delta)

                if terrain == 0:  # Open space
                    if cardinal_delta[0] == 0 and cardinal_delta[1] == 0:
                        self.move(delta)  # Closer than a node distance
                    else:
                        self.move(mathutils.Vector(cardinal_delta))

                    self.check_placables()

                else:  # Something there
                    print("I'm Stuck because I don't know pathfinding")
        if self.selected:
            if self.target_node is not None:
                self.direction_indicator.set_visible(True)
                self.update_direction_indicator()
            else:
                self.direction_indicator.set_visible(False)
        else:
            self.direction_indicator.set_visible(False)


class DirectionIndicator():
    '''Returns an object that can be stretched to be used as a direction
    indicator'''
    def __init__(self):
        scene = bge.logic.getCurrentScene()
        self.obj = scene.addObject('DirectionIndicator', scene.objects[0])
        self.obj.worldOrientation = [0, 0, 0]

    def set_direction(self, from_point, to_point):
        '''Sets the position and direction of the indicator using a
        start and end point'''
        self.obj.worldPosition = from_point
        direction_vect = (from_point - to_point)
        self.obj.worldScale = [1, direction_vect.length, 1]
        self.obj.alignAxisToVect(-direction_vect, 1)
        self.obj.alignAxisToVect([0, 0, 1], 2)
        self.obj.color[0] -= 0.1
        self.obj.color[1] = 1 / direction_vect.length
        self.obj.color[2] = 0.3

    def set_visible(self, state):
        '''Sets visibility of indicator'''
        self.obj.visible = state


def get_nearest_dropsite(pos):
    '''Returns the nearest resource drop point'''
    return Buildings.get_entry().get_map_position()[0:2]


def convert_to_cardinal(direction):
    '''Converts a direction into one of 9 possible directions'''
    direction = mathutils.Vector(direction)
    direction.length = min(1, direction.length * 10)
    out = [round(direction.x), round(direction.y)]
    return out


UNIT_MUTATE_DICT = {
    'Mine': MinerUnit,
    'Resource': ResourceUnit,
}
